@@ -2,6 +2,7 @@
 A custom manager for working with trees of objects.
 """
 from __future__ import unicode_literals
+import functools
 import contextlib
 from itertools import groupby
 
@@ -9,6 +10,7 @@ from django.db import models, connections, router
 from django.db.models import F, ManyToManyField, Max, Q
 from django.utils.translation import ugettext as _
 
+from mptt.compat import remote_field
 from mptt.exceptions import CantDisableUpdates, InvalidMove
 from mptt.querysets import TreeQuerySet
 from mptt.utils import _get_tree_model
@@ -59,7 +61,20 @@ CUMULATIVE_COUNT_SUBQUERY_M2M = """(
 )"""
 
 
+def delegate_manager(method):
+    """
+    Delegate method calls to base manager, if exists.
+    """
+    @functools.wraps(method)
+    def wrapped(self, *args, **kwargs):
+        if self._base_manager:
+            return getattr(self._base_manager, method.__name__)(*args, **kwargs)
+        return method(self, *args, **kwargs)
+    return wrapped
+
+
 class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
+
     """
     A manager for working with trees of objects.
     """
@@ -150,7 +165,7 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             max_attr,
             opts.parent_attr,
             # These fields are used by MPTTModel.update_mptt_cached_fields()
-            *opts.order_insertion_by
+            *[f.lstrip('-') for f in opts.order_insertion_by]
         )
 
         if not q:
@@ -384,25 +399,21 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             new_lookups[join_parts(new_parts)] = v
         return new_lookups
 
+    @delegate_manager
     def _mptt_filter(self, qs=None, **filters):
         """
         Like ``self.filter()``, but translates name-agnostic filters for MPTT
         fields.
         """
-        if self._base_manager:
-            return self._base_manager._mptt_filter(qs=qs, **filters)
-
         if qs is None:
             qs = self
         return qs.filter(**self._translate_lookups(**filters))
 
+    @delegate_manager
     def _mptt_update(self, qs=None, **items):
         """
         Like ``self.update()``, but translates name-agnostic MPTT fields.
         """
-        if self._base_manager:
-            return self._base_manager._mptt_update(qs=qs, **items)
-
         if qs is None:
             qs = self
         return qs.update(**self._translate_lookups(**items))
@@ -472,7 +483,7 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
                     'rel_table': qn(rel_model._meta.db_table),
                     'mptt_fk': qn(rel_model._meta.get_field(rel_field).column),
                     'mptt_table': qn(self.tree_model._meta.db_table),
-                    'mptt_rel_to': qn(mptt_field.rel.field_name),
+                    'mptt_rel_to': qn(remote_field(mptt_field).field_name),
                     'tree_id': qn(meta.get_field(self.tree_id_attr).column),
                     'left': qn(meta.get_field(self.left_attr).column),
                     'right': qn(meta.get_field(self.right_attr).column),
@@ -482,10 +493,11 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
                     'rel_table': qn(rel_model._meta.db_table),
                     'mptt_fk': qn(rel_model._meta.get_field(rel_field).column),
                     'mptt_table': qn(self.tree_model._meta.db_table),
-                    'mptt_rel_to': qn(mptt_field.rel.field_name),
+                    'mptt_rel_to': qn(remote_field(mptt_field).field_name),
                 }
         return queryset.extra(select={count_attr: subquery})
 
+    @delegate_manager
     def insert_node(self, node, target, position='last-child', save=False,
                     allow_existing_pk=False, refresh_target=True):
         """
@@ -506,10 +518,6 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
         set the node's parent and let mptt call this during save.
         """
 
-        if self._base_manager:
-            return self._base_manager.insert_node(
-                node, target, position=position, save=save, allow_existing_pk=allow_existing_pk)
-
         if node.pk and not allow_existing_pk and self.filter(pk=node.pk).exists():
             raise ValueError(_('Cannot insert a node which has already been saved.'))
 
@@ -566,11 +574,8 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             node.save()
         return node
 
+    @delegate_manager
     def _move_node(self, node, target, position='last-child', save=True, refresh_target=True):
-        if self._base_manager:
-            return self._base_manager._move_node(node, target, position=position,
-                                                 save=save, refresh_target=refresh_target)
-
         if self.tree_model._mptt_is_tracking:
             # delegate to insert_node and clean up the gaps later.
             return self.insert_node(node, target, position=position, save=save,
@@ -614,32 +619,25 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
         node_moved.send(sender=node.__class__, instance=node,
                         target=target, position=position)
 
+    @delegate_manager
     def root_node(self, tree_id):
         """
         Returns the root node of the tree with the given id.
         """
-        if self._base_manager:
-            return self._base_manager.root_node(tree_id)
-
         return self._mptt_filter(tree_id=tree_id, parent=None).get()
 
+    @delegate_manager
     def root_nodes(self):
         """
         Creates a ``QuerySet`` containing root nodes.
         """
-        if self._base_manager:
-            return self._base_manager.root_nodes()
-
         return self._mptt_filter(parent=None)
 
+    @delegate_manager
     def rebuild(self):
         """
         Rebuilds all trees in the database table using `parent` link.
         """
-
-        if self._base_manager:
-            return self._base_manager.rebuild()
-
         opts = self.model._mptt_meta
 
         qs = self._mptt_filter(parent=None)
@@ -654,13 +652,12 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             rebuild_helper(pk, 1, idx)
     rebuild.alters_data = True
 
+    @delegate_manager
     def partial_rebuild(self, tree_id):
         """
         Partially rebuilds a tree i.e. It rebuilds only the tree with given
         ``tree_id`` in database table using ``parent`` link.
         """
-        if self._base_manager:
-            return self._base_manager.partial_rebuild(tree_id)
         opts = self.model._mptt_meta
 
         qs = self._mptt_filter(parent=None, tree_id=tree_id)
@@ -847,7 +844,7 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             gap_target_left, gap_size,
             left, right, left_right_change,
             gap_target_left, gap_size,
-            node.pk,
+            node._meta.pk.get_db_prep_value(node.pk, connection),
             getattr(node, self.tree_id_attr)
         ]
         if parent_pk is not None:
@@ -1040,8 +1037,10 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
         # Make space for the subtree which will be moved
         self._create_space(tree_width, space_target, new_tree_id)
         # Move the subtree
+        connection = self._get_connection(instance=node)
         self._inter_tree_move_and_close_gap(
-            node, level_change, left_right_change, new_tree_id, parent.pk)
+            node, level_change, left_right_change, new_tree_id,
+            parent._meta.pk.get_db_prep_value(parent.pk, connection))
 
         # Update the node to be consistent with the updated
         # tree in the database.
@@ -1169,7 +1168,8 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
             left_boundary, right_boundary, gap_size,
             left, right, left_right_change,
             left_boundary, right_boundary, gap_size,
-            node.pk, parent.pk,
+            node._meta.get_field(node._meta.pk.name).get_db_prep_value(node.pk, connection),
+            parent._meta.get_field(parent._meta.pk.name).get_db_prep_value(parent.pk, connection),
             tree_id])
 
         # Update the node to be consistent with the updated
@@ -1236,7 +1236,10 @@ class TreeManager(models.Manager.from_queryset(TreeQuerySet)):
         cursor = connection.cursor()
         cursor.execute(move_tree_query, [
             level_change, left_right_change, left_right_change,
-            new_tree_id, node.pk, parent.pk, left, right, tree_id])
+            new_tree_id,
+            node._meta.pk.get_db_prep_value(node.pk, connection),
+            parent._meta.pk.get_db_prep_value(parent.pk, connection),
+            left, right, tree_id])
 
         # Update the former root node to be consistent with the updated
         # tree in the database.
