@@ -11,6 +11,7 @@ from django.db.models.query_utils import DeferredAttribute
 from django.utils import six
 from django.utils.translation import ugettext as _
 
+from mptt.compat import remote_field
 from mptt.fields import TreeForeignKey, TreeOneToOneField, TreeManyToManyField
 from mptt.managers import TreeManager
 from mptt.signals import node_moved
@@ -131,10 +132,11 @@ class MPTTOptions(object):
                 if f[0] == '-':
                     f = f[1:]
                 field_names.add(f)
+        deferred_fields = instance.get_deferred_fields()
         for field_name in field_names:
-            if instance._deferred:
+            if deferred_fields:
                 field = instance._meta.get_field(field_name)
-                if field.attname in instance.get_deferred_fields() \
+                if field.attname in deferred_fields \
                         and field.attname not in instance.__dict__:
                     # deferred attribute (i.e. via .only() or .defer())
                     # It'd be silly to cache this (that'd do a database query)
@@ -211,7 +213,7 @@ class MPTTOptions(object):
                 # Fall back on tree id ordering if multiple root nodes have
                 # the same values.
                 order_by.append(opts.tree_id_attr)
-            queryset = node.__class__._tree_manager.filter(filters).order_by(*order_by)
+            queryset = node.__class__._tree_manager.db_manager(node._state.db).filter(filters).order_by(*order_by)
             if node.pk:
                 queryset = queryset.exclude(pk=node.pk)
             try:
@@ -334,19 +336,15 @@ class MPTTModelBase(ModelBase):
 
             # Add a tree manager, if there isn't one already
             if not abstract:
-                manager = getattr(cls, 'objects', None)
-                if manager is None:
-                    manager = cls._default_manager._copy_to_model(cls)
-                    manager.contribute_to_class(cls, 'objects')
-                elif manager.model != cls:
-                    # manager was inherited
-                    manager = manager._copy_to_model(cls)
-                    manager.contribute_to_class(cls, 'objects')
-
                 # make sure we have a tree manager somewhere
                 tree_manager = None
-                cls_managers = cls._meta.concrete_managers + cls._meta.abstract_managers
-                for __, __, cls_manager in cls_managers:
+                if hasattr(cls._meta, 'concrete_managers'):  # Django < 1.10
+                    cls_managers = cls._meta.concrete_managers + cls._meta.abstract_managers
+                    cls_managers = [r[2] for r in cls_managers]
+                else:
+                    cls_managers = cls._meta.managers
+
+                for cls_manager in cls_managers:
                     if isinstance(cls_manager, TreeManager):
                         # prefer any locally defined manager (i.e. keep going if not local)
                         if cls_manager.model is cls:
@@ -380,11 +378,12 @@ class MPTTModel(six.with_metaclass(MPTTModelBase, models.Model)):
     """
     Base class for tree models.
     """
-    _default_manager = TreeManager()
 
     class Meta:
         abstract = True
 
+    objects = TreeManager()
+
     def __init__(self, *args, **kwargs):
         super(MPTTModel, self).__init__(*args, **kwargs)
         self._mptt_meta.update_mptt_cached_fields(self)
@@ -782,7 +781,7 @@ class MPTTModel(six.with_metaclass(MPTTModelBase, models.Model)):
         if not self.pk or self._mpttfield('tree_id') is None:
             return False
         opts = self._meta
-        if opts.pk.rel is None:
+        if remote_field(opts.pk) is None:
             return True
         else:
             if not hasattr(self, '_mptt_saved'):
@@ -901,7 +900,11 @@ class MPTTModel(six.with_metaclass(MPTTModelBase, models.Model)):
                         # If we aren't already a descendant of the new parent,
                         # we need to update the parent.rght so things like
                         # get_children and get_descendant_count work correctly.
-                        update_cached_parent = (
+                        #
+                        # parent might be None if parent_id was assigned
+                        # directly -- then we certainly do not have to update
+                        # the cached parent.
+                        update_cached_parent = parent and (
                             getattr(self, opts.tree_id_attr) != getattr(parent, opts.tree_id_attr) or  # noqa
                             getattr(self, opts.left_attr) < getattr(parent, opts.left_attr) or
                             getattr(self, opts.right_attr) > getattr(parent, opts.right_attr))
